/*

   Z----------------------------------Z
   |                                  |
   |    File Created By: Amir Tavafi  |
   |                                  |
   |    Date Created:    2009/11/30,  |
   |                     1388/08/09   |
   |                                  |
   Z----------------------------------Z

*/

#include "bhv_MarlikBlock.h"
#include "bhv_basic_tackle.h"

#include <rcsc/action/body_intercept.h>
#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/soccer_intention.h>
#include <rcsc/player/intercept_table.h>

#include <rcsc/common/server_param.h>

bool Bhv_MarlikBlock::isInBlockPoint = false;
int Bhv_MarlikBlock::timeAtBlockPoint = 0;
rcsc::Vector2D Bhv_MarlikBlock::opp_static_pos = rcsc::Vector2D( -30.0,0.0 );

/// Execute Block Action
bool Bhv_MarlikBlock::execute( rcsc::PlayerAgent * agent )
{
  const rcsc::WorldModel & wm = agent->world();

  rcsc::Vector2D me = wm.self().pos();
  rcsc::Vector2D ball = wm.ball().pos();
  
  int myCycles = wm.interceptTable()->selfReachCycle();
  int tmmCycles = wm.interceptTable()->teammateReachCycle();
  int oppCycles = wm.interceptTable()->opponentReachCycle();

  int num = wm.self().unum();
  
  if( ( wm.self().unum() == 2 && ball.x < -46.0 && ball.y > -18.0 && ball.y < -6.0 &&
        oppCycles <= 3 && oppCycles <= tmmCycles && ball.dist(me) < 9.0 ) ||
      ( wm.self().unum() == 3 && ball.x < -46.0 && ball.y <  18.0 && ball.y >  6.0  &&
        oppCycles <= 3 && oppCycles <= tmmCycles && ball.dist(me) < 9.0 ) )
        if( doBlockMove( agent ) )
          return true;

  rcsc::Vector2D opp = wm.opponentsFromBall().front()->pos();

  float maxDist = 11; // DutchOpen2012: was 14
  
  if( ball.x < -10 )
    maxDist=7; // DutchOpen2012: was 10
  
  if( doInterceptBall2011( agent ) )
     return true;
  if( Bhv_BasicTackle( 0.9, 60.0 ).execute( agent ) )
     return true;

  if( (wm.self().unum() == 2 || wm.self().unum() == 3) && ball.x < -30.0 && ball.absY() > 16.0 )
     return false;

  if( num < 6 && ball.x < -30.0 && ball.x > -40.0 &&
      wm.countTeammatesIn( rcsc::Circle2D( rcsc::Vector2D(opp.x-3.5,opp.y), 3.5 ), 5, false ) > 1 )
    return false;

  if( (wm.self().unum() == 2 || wm.self().unum() == 3) && ball.x > wm.ourDefenseLineX() + maxDist && // for IO2012 changed from + 6.0 to 10.0
      ball.x > -30.0 && ball.x < 0.0 )
     return false;

  if( wm.self().unum() == 6 && ball.x > wm.ourDefenseLineX() + 23.0 )
     return false;

  if( wm.self().unum() == 2 && ball.x < wm.ourDefenseLineX() + 7.0 && ball.absY() < 12.0 &&
      ball.y < wm.self().pos().y + 7.5 && ball.y > wm.self().pos().y - 3.0 && ball.x > -38.0 && oppCycles < tmmCycles )
        if( doBlockMove( agent ) )
          return true;

  if( num == 6 && ball.x < -40.0 && ball.absY() < 6.0 && me.dist(ball) < 5.0 && oppCycles <= tmmCycles  )
        if( doBlockMove( agent ) )
          return true;

  if( wm.self().unum() < 6 && ball.x < wm.ourDefenseLineX() + 5.0 &&
      ball.dist(wm.self().pos()) < 4.0 && ball.x > -38.0 && oppCycles < tmmCycles ) // IO2011: < 6.5
        if( doBlockMove( agent ) )
          return true;

  if( wm.self().unum() == 3 && ball.x < wm.ourDefenseLineX() + 7.0 && ball.absY() < 12.0 &&
      ball.y > wm.self().pos().y + 7.0 && ball.y < wm.self().pos().y + 3.0 && ball.x > -38.0  && oppCycles < tmmCycles )
        if( doBlockMove( agent ) )
          return true;

  if( ( (wm.self().unum() > 5 && wm.self().unum() <= 8 && ball.x < 25.0) ||
      ( (wm.self().unum() <=5 && ball.x < -30.0) ||
        (ball.x < wm.ourDefenseLineX() + maxDist + 2.0 && ball.x > -30.0 && ball.x < 0.0) ) ) && oppCycles < tmmCycles ) // for IO2012 changed from + 8.5 to 12.0
  {

    if( ball.x < -36 && ball.absY() < 7 &&
        myCycles > oppCycles && myCycles <= tmmCycles )
    {
      if( doBlockMove( agent ) )
        return true;
    }
    if( myCycles > oppCycles && myCycles < tmmCycles + 1 )
    {
      if( doBlockMove( agent ) )
        return true;
    }

  }

  isInBlockPoint = false;
  timeAtBlockPoint = 0;

return false;
}

/// Execute Block Action
bool Bhv_MarlikBlock::execute2010( rcsc::PlayerAgent * agent )
{
  const rcsc::WorldModel & wm = agent->world();

  rcsc::Vector2D me = wm.self().pos();
  rcsc::Vector2D ball = wm.ball().pos();

  int myCycles = wm.interceptTable()->selfReachCycle();
  int tmmCycles = wm.interceptTable()->teammateReachCycle();
  int oppCycles = wm.interceptTable()->opponentReachCycle();

  if( ( wm.self().unum() == 2 && ball.x < -46.0 && ball.y > -18.0 && ball.y < -6.0 &&
        oppCycles <= 3 && oppCycles <= tmmCycles && ball.dist(me) < 9.0 ) ||
      ( wm.self().unum() == 3 && ball.x < -46.0 && ball.y <  18.0 && ball.y >  6.0  &&
        oppCycles <= 3 && oppCycles <= tmmCycles && ball.dist(me) < 9.0 ) )
        if( doBlockMove( agent ) )
          return true;

  if( doInterceptBall( agent ) )
     return true;
  if( Bhv_BasicTackle( 0.85, 60.0 ).execute( agent ) )
     return true;

  if( (wm.self().unum() == 2 || wm.self().unum() == 3) && ball.x < -30.0 && ball.absY() > 18.0 )
     return false;

  if( (wm.self().unum() == 2 || wm.self().unum() == 3) && ball.x > wm.ourDefenseLineX() + 6.0 &&
      ball.x > -30.0 && ball.x < 0.0 )
     return false;

  if( wm.self().unum() == 6 && ball.x > wm.ourDefenseLineX() + 15.0 )
     return false;

  if( wm.self().unum() == 2 && ball.x < wm.ourDefenseLineX() + 7.0 && ball.absY() < 12.0 &&
      ball.y < wm.self().pos().y + 9.0 && ball.y > wm.self().pos().y - 3.0 && ball.x > -38.0 )
        if( doBlockMove( agent ) )
          return true;

  if( wm.self().unum() < 6 && ball.x < wm.ourDefenseLineX() + 5.0 &&
      ball.dist(wm.self().pos()) < 5.0 && ball.x > -38.0 ) // IO2011: < 6.5
        if( doBlockMove( agent ) )
          return true;

  if( wm.self().unum() == 3 && ball.x < wm.ourDefenseLineX() + 7.0 && ball.absY() < 12.0 &&
      ball.y > wm.self().pos().y + 7.0 && ball.y < wm.self().pos().y + 3.0 && ball.x > -38.0 )
        if( doBlockMove( agent ) )
          return true;

  if( (wm.self().unum() > 5 && wm.self().unum() <= 8 && ball.x < 25.0) ||
      ( (wm.self().unum() <=5 && ball.x < -30.0) ||
        (ball.x < wm.ourDefenseLineX() + 8.5 && ball.x > -30.0 && ball.x < -5.0) ) )
  {

    if( ball.x < -36 && ball.absY() < 7 &&
        myCycles > oppCycles && myCycles <= tmmCycles && oppCycles < tmmCycles )
    {
      if( doBlockMove( agent ) )
        return true;
    }
    if( myCycles > oppCycles && myCycles < tmmCycles + 1 && oppCycles < tmmCycles )
    {
      if( doBlockMove( agent ) )
        return true;
    }

  }


return false;
}

/// Do Intercept if Possible
bool Bhv_MarlikBlock::doInterceptBall( rcsc::PlayerAgent * agent )
{

    const rcsc::WorldModel & wm = agent->world();

    rcsc::Vector2D ball = wm.ball().pos();
    rcsc::Vector2D me = wm.self().pos();

    int myCycles = wm.interceptTable()->selfReachCycle();
    int tmmCycles = wm.interceptTable()->teammateReachCycle();
    int oppCycles = wm.interceptTable()->opponentReachCycle();

    if ( myCycles < oppCycles && myCycles < tmmCycles )
    {
        rcsc::Body_Intercept(true,ball).execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        return true;
    }

    rcsc::Vector2D myInterceptPos = wm.ball().inertiaPoint( myCycles );
    rcsc::Vector2D opp_reach_pos = wm.ball().inertiaPoint( oppCycles );

    if ( ( (myCycles <= 4 && oppCycles >= 4) || myCycles <= 2 ) &&
         me.x < myInterceptPos.x && me.x < ball.x - 0.5 &&
         ( std::fabs( ball.y - me.y ) < 1.5 || ball.absY() < me.absY() ) )
    {
        rcsc::Body_Intercept(true,ball).execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        return true;
    }


return false;
}

/// Do Intercept if Possible
bool Bhv_MarlikBlock::doInterceptBall2011( rcsc::PlayerAgent * agent )
{

    const rcsc::WorldModel & wm = agent->world();

    rcsc::Vector2D ball = wm.ball().pos();
    rcsc::Vector2D me = wm.self().pos();

    int myCycles = wm.interceptTable()->selfReachCycle();
    int tmmCycles = wm.interceptTable()->teammateReachCycle();
    int oppCycles = wm.interceptTable()->opponentReachCycle();

    if ( myCycles < oppCycles && myCycles < tmmCycles )
    {
        rcsc::Body_Intercept(true,ball).execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        return true;
    }

    rcsc::Vector2D myInterceptPos = wm.ball().inertiaPoint( myCycles );
    rcsc::Vector2D opp_reach_pos = wm.ball().inertiaPoint( oppCycles );

/*
    if ( ( (myCycles <= 4 && oppCycles >= 4) || myCycles <= 2 ) &&
         me.x < myInterceptPos.x && me.x < ball.x - 0.5 &&
         ( std::fabs( ball.y - me.y ) < 1.5 || ball.absY() < me.absY() ) )
    {
        rcsc::Body_Intercept().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        return true;
    }
*/

return false;
}

/// Do Block when in front of opponent for near distances AmirZ::RC2012
bool Bhv_MarlikBlock::doBlockNear2012(rcsc::PlayerAgent* agent)
{
  const rcsc::WorldModel & wm = agent->world();

  rcsc::Vector2D ball = wm.ball().pos();
  rcsc::Vector2D me = wm.self().pos() + wm.self().vel();

  int myCycles = wm.interceptTable()->selfReachCycle();
  int tmmCycles = wm.interceptTable()->teammateReachCycle();
  int oppCycles = wm.interceptTable()->opponentReachCycle();

  rcsc::Vector2D blockPos = getBlockPoint( agent );
  rcsc::Vector2D opp = wm.opponentsFromBall().front()->pos() + wm.opponentsFromBall().front()->vel();
  rcsc::Vector2D oppNoVel = wm.opponentsFromBall().front()->pos();
  rcsc::Vector2D goal = rcsc::Vector2D(-52.0, 0.0);

  double dashPower = getBlockDashPower( agent, blockPos );

  rcsc::Vector2D oppPos = wm.opponentsFromBall().front()->pos();
  rcsc::Vector2D oppVel = wm.opponentsFromBall().front()->vel();

  rcsc::Vector2D oppTarget = getOppTarget(agent,ball);

  if( oppTarget.x < -999.0 )
    return false;

  rcsc::Vector2D stopPoint = getStopPoint(agent);

  if( me.dist(blockPos) < 5.0 && me.dist(goal) < oppPos.dist(goal) )
  {
    
    
  }


return true;
}

/// get stop position for near block
rcsc::Vector2D Bhv_MarlikBlock::getStopPoint(rcsc::PlayerAgent* agent)
{
  const rcsc::WorldModel & wm = agent->world();

  rcsc::Vector2D ball = wm.ball().pos();
  rcsc::Vector2D me = wm.self().pos();
  rcsc::Vector2D blockPos = ball;

//   int myCycles = wm.interceptTable()->selfReachCycle();
//   int tmmCycles = wm.interceptTable()->teammateReachCycle();
  int oppCycles = wm.interceptTable()->opponentReachCycle();

  float oppDribbleSpeed = 0.675; // 0.75 bud!

  rcsc::Vector2D ballPred = ball;

   if( !wm.existKickableOpponent() )
        ballPred = wm.ball().inertiaPoint( oppCycles );

  rcsc::Vector2D oppPos = wm.opponentsFromBall().front()->pos();
  rcsc::Vector2D oppVel = wm.opponentsFromBall().front()->vel();
  rcsc::Vector2D oppTarget = getOppTarget(agent,ball);
          
  rcsc::Vector2D nextBallPos = ballPred;
  nextBallPos += rcsc::Vector2D::polar2vector( oppDribbleSpeed, (oppTarget - ball).dir() );

  int oppTime;
   if( wm.existKickableOpponent() )
      oppTime = 0;
   else
      oppTime = oppCycles;

  while( wm.self().playerType().cyclesToReachDistance( me.dist(nextBallPos) ) > oppTime && oppTime < 50 )
  {
       if( nextBallPos.x > 36.0 && (wm.self().unum() == 7 || wm.self().unum() == 9) )
        oppTarget = rcsc::Vector2D( -36.0, -12.0 );
       if( nextBallPos.x > 30.0 && wm.self().unum() == 11 )
        oppTarget = rcsc::Vector2D( -30.0, 0.0 );
       if( nextBallPos.x < -34.0 && nextBallPos.x > -40.0 )
        oppTarget = rcsc::Vector2D( -48.0, 0.0 );
       if( nextBallPos.x < -40.0 )
        oppTarget = rcsc::Vector2D( -48.0, 0.0 );
    
       nextBallPos += rcsc::Vector2D::polar2vector( oppDribbleSpeed, (oppTarget - ball).dir() );
       oppTime++;
  }

  if( oppTime >= 50 )
       blockPos = rcsc::Vector2D( -48.0, ball.y );
  else
       blockPos = nextBallPos;

  if( wm.self().unum() < 6 && blockPos.x > 0.0 )
       blockPos.x = 10.0;

return blockPos;  
}

/// To Get Opponent Target
rcsc::Vector2D Bhv_MarlikBlock::getOppTarget(rcsc::PlayerAgent* agent, rcsc::Vector2D ball)
{
  const rcsc::WorldModel & wm = agent->world();

  rcsc::Vector2D me = wm.self().pos() + wm.self().vel();

  int myCycles = wm.interceptTable()->selfReachCycle();
  int tmmCycles = wm.interceptTable()->teammateReachCycle();
  int oppCycles = wm.interceptTable()->opponentReachCycle();
  
  if( oppCycles > 0 )
    ball = wm.ball().inertiaPoint(oppCycles);
  
  if( ! wm.opponentsFromBall().front()->pos().isValid() ||  ! wm.opponentsFromBall().front()->vel().isValid() )
    return rcsc::Vector2D(-1000.0,-1000.0);

  rcsc::Vector2D blockPos = getBlockPoint( agent );
  rcsc::Vector2D oppNoVel = wm.opponentsFromBall().front()->pos();
  rcsc::Vector2D goal = rcsc::Vector2D(-52.0, 0.0);

  double dashPower = getBlockDashPower( agent, blockPos );

  rcsc::Vector2D oppPos = wm.opponentsFromBall().front()->pos();
  rcsc::Vector2D oppVel = wm.opponentsFromBall().front()->vel();
  rcsc::Vector2D opp = oppPos + oppVel;
  rcsc::Vector2D target(-1000.0,-1000.0);
  
  
  if( oppPos.x < -36.0 )
  {
    if( oppPos.absY() < 8.0 )
      target = rcsc::Vector2D( oppPos.x-10,oppPos.y*0.8 );
    else
      target = rcsc::Vector2D(-49.0,0.0);
  }
  else if( oppPos.x < -30 )
    target = rcsc::Vector2D( oppPos.x-10,oppPos.y*0.875 );
  else
    target = rcsc::Vector2D( oppPos.x-20, oppPos.y*0.96 );
  
return target;
}

/// Going to Block Point & related Body Turns
bool Bhv_MarlikBlock::doBlockMove( rcsc::PlayerAgent * agent )
{
  const rcsc::WorldModel & wm = agent->world();

  static bool isChaporasting;
  
  rcsc::Vector2D ball = wm.ball().pos();
  rcsc::Vector2D me = wm.self().pos() + wm.self().vel();

  int myCycles = wm.interceptTable()->selfReachCycle();
  int tmmCycles = wm.interceptTable()->teammateReachCycle();
  int oppCycles = wm.interceptTable()->opponentReachCycle();

  rcsc::Vector2D blockPos = getBlockPoint( agent );
  rcsc::Vector2D opp = wm.opponentsFromBall().front()->pos() + wm.opponentsFromBall().front()->vel();
  rcsc::Vector2D oppNoVel = wm.opponentsFromBall().front()->pos();
  rcsc::Vector2D goal = rcsc::Vector2D(-52.0, 0.0);

  double dashPower = getBlockDashPower( agent, blockPos );

  rcsc::Vector2D oppPos = wm.opponentsFromBall().front()->pos();
  rcsc::Vector2D oppVel = wm.opponentsFromBall().front()->vel();
  
  rcsc::Vector2D oppTarget = rcsc::Vector2D( -52.5, oppPos.y*0.95 );

   if( ball.x > 36.0 && (wm.self().unum() == 7 || wm.self().unum() == 9) )
        oppTarget = rcsc::Vector2D( -36.0, -12.0 );
   if( wm.self().unum() == 8 || wm.self().unum() == 10 )
        oppTarget = rcsc::Vector2D( -36.0, 12.0 );

   if( ball.x > 30.0 && wm.self().unum() == 11 )
        oppTarget = rcsc::Vector2D( -30.0, 0.0 );

   if( ball.x < -34.0 && ball.x > -40.0 )
        oppTarget = rcsc::Vector2D( -48.0, 0.0 );
   
   if( ball.x < -30.0 && ball.x > -40.0 && ball.absY() < 15 )
   {
     if( wm.self().unum() == 2 )
        oppTarget = rcsc::Vector2D( -52.5, -4.0 );
     if( wm.self().unum() == 3 )
        oppTarget = rcsc::Vector2D( -52.5, 4.0 );
	
   }
   
   if( ball.dist(goal) < 14 && ball.absY() > 8 )
        oppTarget = rcsc::Vector2D( -51.0, 0.0 );

   if( ball.x < -40.0 )
        oppTarget = rcsc::Vector2D( -49.0, 0.0 );
   
   if( ball.x < -47.0 )
        oppTarget = rcsc::Vector2D( -50.0, 0.0 );

   rcsc::Line2D targetLine = rcsc::Line2D( oppTarget, oppPos+oppVel );
     
   if( targetLine.dist(opp_static_pos) >= 0.2 )
   {
      if( oppVel.y > 0.0 )
         oppTarget = rcsc::Vector2D(-52.5, 9);
      if( oppVel.y < 0.0 )
         oppTarget = rcsc::Vector2D(-52.5, -9);
   }


  const rcsc::PlayerObject * opponent = wm.interceptTable()->fastestOpponent();
  rcsc::AngleDeg opp_body;


   if( ball.x < -36.0 && ball.absY() > 12.0 && me.dist(goal) > opp.dist(goal) )
     return false;


   int minOppDist = 3.5;

    if( ball.x < -36.0 && ball.absY() < 12.0 )
         minOppDist = 4.0;


  int minInterceptCycles = 2;

    if( ball.x < -36.0 && ball.absY() > 12.0 && me.dist(goal) < opp.dist(goal) )
         minInterceptCycles = 3; // 3 bud!
    if( ball.x < -36.0 && ball.absY() < 13.0 )
         minInterceptCycles = 4;
    if( ball.dist(goal) < 13 )
         minInterceptCycles = 5;

    if( ball.x < -36.0 && wm.self().unum() == 6 )
      minInterceptCycles = 5;

    if( wm.self().unum() == 7 || wm.self().unum() == 8 )
      minInterceptCycles = 7;
    
    
     rcsc::Vector2D oppTarget22 = rcsc::Vector2D( -52.5, oppPos.y*0.95 ); // oppPos.y*0.975

     if( ball.x > 36.0 && (wm.self().unum() == 7 || wm.self().unum() == 9) )
        oppTarget22 = rcsc::Vector2D( -36.0, -12.0 );
     if( wm.self().unum() == 8 || wm.self().unum() == 10 )
        oppTarget22 = rcsc::Vector2D( -36.0, 12.0 );

     if( ball.x > 30.0 && wm.self().unum() == 11 )
        oppTarget22 = rcsc::Vector2D( -30.0, 0.0 );

     if( ball.x < -34.0 && ball.x > -40.0 )
        oppTarget22 = rcsc::Vector2D( -48.0, 0.0 );

     if( ball.x < -30.0 && ball.x > -40.0 && ball.absY() < 15 )
     {
       if( wm.self().unum() == 2 )
          oppTarget22 = rcsc::Vector2D( -52.5, -4.0 );
       if( wm.self().unum() == 3 )
          oppTarget22 = rcsc::Vector2D( -52.5, 4.0 );	
     }
     
     if( ball.x < -40.0 )
        oppTarget22 = rcsc::Vector2D( -49.0, 0.0 );

     if( ball.x < -47.0 )
        oppTarget22 = rcsc::Vector2D( -50.0, 0.0 );


  static rcsc::Vector2D opp_static_pos = opp;

    if( myCycles <= 1 )
    {
      isChaporasting = false;
      rcsc::Body_Intercept(true,ball).execute( agent );
      agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
      return true;
    }
    if( myCycles < 20 && ( myCycles < tmmCycles || ( myCycles < tmmCycles + 2 && tmmCycles > 3 ) ) &&
         myCycles <= oppCycles )
    {
      isChaporasting = false;
      rcsc::Body_Intercept(true,ball).execute( agent );
      agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
      return true;
    }
    if( ball.x < -49.0 && ( (ball.y < 14 && ball.y > 7.0) || (ball.y > -14 && ball.y < -7.0) ) && myCycles < 5 ) // AmirZ RC2012
    {
      isChaporasting = false;
      rcsc::Body_Intercept(true,ball).execute( agent );
      agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
      return true;
    }
    if( wm.self().unum() > 5 && ball.dist(goal) < 13.0 && myCycles < 5 && me.dist(goal) < ball.dist(goal) ) // AmirZ RC2012
    {
      isChaporasting = false;
      rcsc::Body_Intercept(true,ball).execute( agent );
      agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
      return true;
    }


    if( myCycles < tmmCycles && oppCycles >= 2 && myCycles <= oppCycles  )
    {
      isChaporasting = false;
      rcsc::Body_Intercept(true,ball).execute( agent );
      agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
      return true;
    }


   if( me.dist( blockPos ) < 1.7 && targetLine.dist(me) < 1.25 && myCycles <= minInterceptCycles &&
       targetLine.dist(opp_static_pos) < 0.2 )
   {
      isChaporasting = false;
      timeAtBlockPoint = 0;
      rcsc::Body_Intercept(true,ball).execute( agent );
      agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
      return true;
   }
   else if( (wm.self().unum()==7 || wm.self().unum()==8) && me.dist( blockPos ) < 3.0 && targetLine.dist(me) < 2 &&
            myCycles <= minInterceptCycles && targetLine.dist(opp_static_pos) < 0.5 )
   {
      isChaporasting = false;
      timeAtBlockPoint = 0;
      rcsc::Body_Intercept(true,ball).execute( agent );
      agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
      return true;
   }
   else if( ball.x < -36.0 && ball.absY() > 14.0 && me.dist(goal) + 1.5 < opp.dist(goal) && targetLine.dist(me) < 2.0 &&
            me.dist(opp) < 5.0 && me.dist(blockPos) < 3.0 && oppCycles < 3 && targetLine.dist(opp_static_pos) < 0.5 )
   // RC2012: for cross area
   {
      isChaporasting = false;
      timeAtBlockPoint = 0;
      rcsc::Body_Intercept(true,ball).execute( agent );
      agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
      std::cout<<"\n\nxXxXxXxXxXxXxXx CYCLE "<<wm.time().cycle()<<" Block Intercepting while cross 2012 xXxXxXxXx\n\n";
      return true;
   }
   else if( (me.dist( blockPos ) < 3.5 || me.dist(opp) < 2.5 ) && me.dist( opp ) < minOppDist && ball.dist( oppNoVel ) < 1.2 &&
            me.dist(goal)+0.8 < opp.dist(goal) && targetLine.dist(me) < 0.9 &&
            targetLine.dist(opp_static_pos) < 0.2 ) // vele harif age tu masire targetesh nabud, nare tu in if!
   {
     
      if( !rcsc::Body_GoToPoint( opp, 0.4, dashPower, 1, false, true ).execute( agent ) ) // dist_thr 0.5 bud
      {
         rcsc::AngleDeg bodyAngle = agent->world().ball().angleFromSelf();
         if( me.y < 0.0 )
                bodyAngle -= 90.0;
         else
                bodyAngle += 90.0;

         rcsc::Body_TurnToAngle( bodyAngle ).execute( agent );
      }
     isChaporasting = false;

   }
   else if( (me.dist( blockPos ) < 4.0 || me.dist(opp) < 2.2 ) && me.dist( opp ) < 4.0 && ball.dist( oppNoVel ) < 1.2 &&
            me.dist(goal)+0.8 < opp.dist(goal) && wm.opponentsFromBall().front()->vel().r() < 0.1 &&
            targetLine.dist(me) < 0.9 )
   {
     
      if( !rcsc::Body_GoToPoint( opp, 0.4, dashPower, 1, false, true ).execute( agent ) ) // dist_thr 0.5 bud
      {
         rcsc::AngleDeg bodyAngle = agent->world().ball().angleFromSelf();
         if( me.y < 0.0 )
                bodyAngle -= 90.0;
         else
                bodyAngle += 90.0;

         rcsc::Body_TurnToAngle( bodyAngle ).execute( agent );
      }
     isChaporasting = false;

   }
   else if( (me.dist(blockPos) < 2.0 && me.dist(goal) < opp.dist(goal) && ball.dist( opp ) < 1.3
            /*&& timeAtBlockPoint > 0*/) || isChaporasting )
   {
     if( me.dist(goal) + 1 > opp.dist(goal) )
       isChaporasting = false;
     
     if( me.dist(goal) + 1 < blockPos.dist(goal) )
       isChaporasting = false;
     
     if( ball.dist(oppNoVel) > 1.1 )
       isChaporasting = false;

     if( me.dist(blockPos) > 4.0 )
       isChaporasting = false;
    
     if( ball.x < -37.0 && ball.absY() < 10.0 && me.dist(blockPos) < 3.0 )
       isChaporasting = false;

     if( opp.dist(me) > 3.0 )
       isChaporasting = false;
     
     double power = 100;

      if( opp.dist(opp_static_pos) > 0.3 )
        power = 100;
      else if( opp.dist(opp_static_pos) > 0.2 )
        power = 80;
      else
	power = 60;
     
//      rcsc::AngleDeg bodyAngle = agent->world().ball().angleFromSelf();
     rcsc::AngleDeg bodyAngle = (opp-oppTarget22).dir();
	
     if( me.y < 0 )
        bodyAngle -= 90.0;
     else
        bodyAngle += 90.0;

      
     if( std::fabs(wm.self().body().degree() - bodyAngle.degree()) < 10 )
     {
      rcsc::Line2D oppLine(opp,oppTarget22);

      float myXOnLine= oppLine.getX(me.y);
      float myYOnLine= oppLine.getY(me.x);
       
      if( ball.x < -30 && ball.absY() > 15 )
      {

       if( me.y > 0 )
       {
	if( wm.self().stamina()>4000 )
	{
	 if( me.x < myXOnLine )
	   agent->doDash(-power);
	 else
	   agent->doDash(power);
	}
/*	 std::cout<<"\n***** CYCLE:  "<<wm.time().cycle()<<" No "<<wm.self().unum()<<" Target: "
	 <<oppTarget22<<" \n";  */
       }
       else
       {
	if( wm.self().stamina()>4000 )
	{
	 if( me.x > myXOnLine )
	   agent->doDash(power);
	 else
	   agent->doDash(-power);
	}
/*	 std::cout<<"\n***** CYCLE:  "<<wm.time().cycle()<<" No "<<wm.self().unum()<<" Target: "
	 <<oppTarget22<<" \n";*/
       }
       
      }
      else
      {

       if( me.y > 0 )
       {
	if( wm.self().stamina()>4000 )
	{
	 if( me.y < myYOnLine )
	   agent->doDash(power);
	 else
	   agent->doDash(-power);
	}
/*	 std::cout<<"\n***** CYCLE:  "<<wm.time().cycle()<<" No "<<wm.self().unum()<<" Target: "
	 <<oppTarget22<<" \n";  */
       }
       else
       {
	if( wm.self().stamina()>4000 )
	{
	 if( me.y > myYOnLine )
	   agent->doDash(power);
	 else
	   agent->doDash(-power);
	}
/*	 std::cout<<"\n***** CYCLE:  "<<wm.time().cycle()<<" No "<<wm.self().unum()<<" Target: "
	 <<oppTarget22<<" \n";*/
       }
 
      }
      
       timeAtBlockPoint++;
     }
     else
     {
       rcsc::Body_TurnToAngle( bodyAngle ).execute( agent );
       timeAtBlockPoint++;
     }
    
     
     isChaporasting = true;
     //std::cout<<"\n******** CYCLE: "<<wm.time().cycle()<<" No. "<<wm.self().unum()<<" isChaporasting = TRUE \n";
     
   }   
   else if( !rcsc::Body_GoToPoint( blockPos, 0.4, dashPower, 1, 100, true, 30.0 ).execute( agent ) ) // dist_thr 0.5 bud
   {
//        rcsc::AngleDeg bodyAngle = agent->world().ball().angleFromSelf();
// 
//        if ( bodyAngle.degree() < 0.0 )
//               bodyAngle -= 90.0;
//        else
//               bodyAngle += 90.0;

     rcsc::AngleDeg bodyAngle = (opp-oppTarget22).dir();

     if( me.y < 0 )
        bodyAngle -= 90.0;
     else
        bodyAngle += 90.0;


       rcsc::Body_TurnToAngle( bodyAngle ).execute( agent );
       timeAtBlockPoint++;
       isChaporasting=false;
   }
   else
     timeAtBlockPoint = 0;

   opp_static_pos = opp;

   isChaporasting=false;

   if ( wm.ball().distFromSelf() < 20.0 &&
        ( wm.existKickableOpponent() || oppCycles <= 3 ) )
       agent->setNeckAction( new rcsc::Neck_TurnToBall() );
   else
       agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );

return true;
}

/// get Opponent BlockPoint for Moving to
rcsc::Vector2D Bhv_MarlikBlock::getBlockPoint( rcsc::PlayerAgent * agent )
{
  const rcsc::WorldModel & wm = agent->world();

  rcsc::Vector2D ball = wm.ball().pos();
  rcsc::Vector2D me = wm.self().pos();
  rcsc::Vector2D blockPos = ball;

//   int myCycles = wm.interceptTable()->selfReachCycle();
//   int tmmCycles = wm.interceptTable()->teammateReachCycle();
  int oppCycles = wm.interceptTable()->opponentReachCycle();

  float oppDribbleSpeed = 0.675; // 0.75 bud!

  rcsc::Vector2D ballPred = ball;

   if( !wm.existKickableOpponent() )
        ballPred = wm.ball().inertiaPoint( oppCycles );

  rcsc::Vector2D oppPos = wm.opponentsFromBall().front()->pos();
  rcsc::Vector2D oppVel = wm.opponentsFromBall().front()->vel();

  rcsc::Vector2D oppTarget = rcsc::Vector2D( -52.5, oppPos.y*0.95 ); // oppPos.y*0.975

     if( ball.x > 36.0 && (wm.self().unum() == 7 || wm.self().unum() == 9) )
        oppTarget = rcsc::Vector2D( -36.0, -12.0 );
     if( wm.self().unum() == 8 || wm.self().unum() == 10 )
        oppTarget = rcsc::Vector2D( -36.0, 12.0 );

     if( ball.x > 30.0 && wm.self().unum() == 11 )
        oppTarget = rcsc::Vector2D( -30.0, 0.0 );

     if( ball.x < -34.0 && ball.x > -40.0 )
        oppTarget = rcsc::Vector2D( -48.0, 0.0 );

     if( ball.x < -30.0 && ball.x > -40.0 && ball.absY() < 15 )
     {
       if( wm.self().unum() == 2 )
          oppTarget = rcsc::Vector2D( -52.5, -4.0 );
       if( wm.self().unum() == 3 )
          oppTarget = rcsc::Vector2D( -52.5, 4.0 );	
     }

     if( ball.dist(rcsc::Vector2D(-52.5,0.0)) < 14 && ball.absY() > 8 )
        oppTarget = rcsc::Vector2D( -51.0, 0.0 );

     if( ball.x < -40.0 )
        oppTarget = rcsc::Vector2D( -49.0, 0.0 );

     if( ball.x < -47.0 )
        oppTarget = rcsc::Vector2D( -50.0, 0.0 );
    
   rcsc::Vector2D nextBallPos = ballPred;
   nextBallPos += rcsc::Vector2D::polar2vector( oppDribbleSpeed, (oppTarget - ballPred).dir() );

  int oppTime;
   if( wm.existKickableOpponent() )
      oppTime = 0;
   else
      oppTime = oppCycles;

  while( wm.self().playerType().cyclesToReachDistance( me.dist(nextBallPos) ) > oppTime && oppTime < 50 )
  {
       if( nextBallPos.x > 36.0 && (wm.self().unum() == 7 || wm.self().unum() == 9) )
        oppTarget = rcsc::Vector2D( -36.0, -12.0 );
       if( nextBallPos.x > 30.0 && wm.self().unum() == 11 )
        oppTarget = rcsc::Vector2D( -30.0, 0.0 );
       if( nextBallPos.x < -34.0 && nextBallPos.x > -40.0 )
        oppTarget = rcsc::Vector2D( -48.0, 0.0 );
       if( nextBallPos.x < -40.0 )
        oppTarget = rcsc::Vector2D( -48.0, 0.0 );
    
       nextBallPos += rcsc::Vector2D::polar2vector( oppDribbleSpeed, (oppTarget - ballPred).dir() );
       oppTime++;
  }

  if( oppTime >= 50 )
       blockPos = rcsc::Vector2D( -48.0, ball.y );
  else
       blockPos = nextBallPos;

  if( wm.self().unum() < 6 && blockPos.x > 0.0 )
       blockPos.x = 10.0;

return blockPos;
}

/// get Dash Power for Move
double Bhv_MarlikBlock::getBlockDashPower( const rcsc::PlayerAgent * agent,
                             const rcsc::Vector2D & blockPos )
{

    const rcsc::WorldModel & wm = agent->world();

    double dashPower = rcsc::ServerParam::i().maxDashPower();
//     const double my_inc = wm.self().playerType().staminaIncMax() * wm.self().recovery();

    int myCycles = wm.interceptTable()->selfReachCycle();
    int tmmCycles = wm.interceptTable()->teammateReachCycle();
    int oppCycles = wm.interceptTable()->opponentReachCycle();

    if ( wm.ourDefenseLineX() > wm.self().pos().x
         && wm.ball().pos().x < wm.ourDefenseLineX() + 20.0 )
    {
        dashPower = rcsc::ServerParam::i().maxDashPower();
    }
    else if ( !wm.existKickableTeammate()
              && oppCycles < myCycles
              && oppCycles < tmmCycles
              && blockPos.x < wm.self().pos().x
              && blockPos.x < -25.0
              && wm.ball().inertiaPoint( oppCycles ).x < -25.0 )
    {
        dashPower = rcsc::ServerParam::i().maxDashPower();
    }
    else
    {
        dashPower = rcsc::ServerParam::i().maxDashPower();
    }

return dashPower;
}

